/*
 * Copyright (c) 2011-2014 Nokia Solutions and Networks
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * @file fzm_cpld.h
 * @author Nokia Solutions and Networks
 * @remarks Copyright (C) 2011-2014 Nokia Solutions and Networks
 *
 * @par
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * @par PURPOSE
 * FZM platform specific CPLD header file.
 *
 * @par REVISIONS
 * <pre>
 * Date       CoreID    MOTCM      Description
 * ---------- --------- ---------- -------------------------------------------
 * 07/26/2011 qdpx86               Initial creation
 * 01/31/2012 tberger1             Updated for krypton
 * 04/11/2013 bbausch1             Porting to FZM
 * </pre>
 */
#ifndef FZM_CPLD_H
#define FZM_CPLD_H

#include <linux/types.h>

#ifdef __KERNEL__
#  include "fzm_cpldspi.h"
#endif

#define CPLD_SIZE (1<<10)

struct cpld_regs {
#define REG(x) __u8 x; __u8 __pad_##x[3];
#define REGW(x) REG(x)
#define REGR(x) REG(x)
#define REGG(x) REG(x)
   /* really weird hack to get __COUNTER__ to expand properly */
#define RESVTMP2(n,tmp) __u8 __pad ## tmp [(n)*4];
#define RESVTMP(n,tmp) RESVTMP2(n,tmp)
#define RESV(n) RESVTMP(n, __COUNTER__)
#  include "fzm_cpld_regs.def"
#undef RESVTMP
#undef RESVTMP2
};

/****************************************************************************
 * FZM specific CPLD register bits / masks
 ***************************************************************************/
#define CPLD_RSR1_BIT_PU         0x01
#define CPLD_RSR1_BIT_HARD       0x02
#define CPLD_RSR1_BIT_FR         0x04
#define CPLD_RSR1_BIT_L2WD       0x08
#define CPLD_RSR1_BIT_RFD        0x10
#define CPLD_RSR1_BIT_SOC        0x20
#define CPLD_RSR1_BIT_TEMP       0x40
#define CPLD_RSR1_BIT_EMU        0x80

/* needed ? */
#define CPLD_RSR2_BIT_RFD        0x01
#define CPLD_RSR2_BIT_RFD_WIFI   0x02

#define CPLD_RRR1_BIT_HARD       0x02
#define CPLD_RRR1_BIT_FULL       0x04
#define CPLD_RRR1_BIT_POR        0x20

#define CPLD_RRR2_BIT_BHPHY      0x01
#define CPLD_RRR2_BIT_BHPHY2     0x02
#define CPLD_RRR2_BIT_BHPHY3     0x08
#define CPLD_RRR2_BIT_FPGA       0x10
#define CPLD_RRR2_BIT_PSE        0x20
#define CPLD_RRR2_BIT_EXT_SWITCH 0x40
#define CPLD_RRR2_BIT_EXT_WIFI   0x80

#define CPLD_PSTKYREN1_BIT_BHPHY 0x01
#define CPLD_PSTKYREN1_BIT_BHPHY2 0x02
#define CPLD_PSTKYREN1_BIT_BHPHY3 0x08
#define CPLD_PSTKYREN1_BIT_FPGA  0x10
#define CPLD_PSTKYREN1_BIT_PSE   0x20
#define CPLD_PSTKYREN1_BIT_EXT_SWITCH 0x40

#define CPLD_PSTKYREN2_BIT_TXDIS1 0x01
#define CPLD_PSTKYREN2_BIT_TXDIS2 0x40

#define CPLD_WCTR_BIT_FCTRY_RST_WIFI  0x01
#define CPLD_WCTR_BIT_SW_RST_WIFI     0x02
#define CPLD_WCTR_BIT_WIFI_PWR        0x10
#define CPLD_WCTR_BIT_SW_CTL_WIFI_LED 0x80

#define CPLD_SFPC_BIT_TXDIS      0x01
#define CPLD_SFPC_BIT_RTSEL      0x02
#define CPLD_SFPC_BIT_TXDIS_2    0x04
#define CPLD_SFPC_BIT_RTSEL_2    0x08

#define CPLD_LED1_BIT_RED1       0x01
#define CPLD_LED1_BIT_GRN1       0x02
#define CPLD_LED1_BIT_RED2       0x04
#define CPLD_LED1_BIT_GRN2       0x08
#define CPLD_LED1_BIT_RED3       0x10
#define CPLD_LED1_BIT_GRN3       0x20
#define CPLD_LED1_BIT_DIS        0x80

#define CPLD_LED2_MASK_CTR1      0x03
#define CPLD_LED2_SHIFT_CTR1     0
#define CPLD_LED2_SET_CTR1(x)    (((x)<<CPLD_LED2_SHIFT_CTR1) & CPLD_LED2_MASK_CTR1)
#define CPLD_LED2_MASK_CTR2      0x0c
#define CPLD_LED2_SHIFT_CTR2     2
#define CPLD_LED2_SET_CTR2(x)    (((x)<<CPLD_LED2_SHIFT_CTR2) & CPLD_LED2_MASK_CTR2)
#define CPLD_LED2_MASK_CTR3      0x30
#define CPLD_LED2_SHIFT_CTR3     4
#define CPLD_LED2_SET_CTR3(x)    (((x)<<CPLD_LED2_SHIFT_CTR3) & CPLD_LED2_MASK_CTR3)
#define CPLD_LED2_MASK_CTR4      0xc0
#define CPLD_LED2_SHIFT_CTR4     6
#define CPLD_LED2_SET_CTR4(x)    (((x)<<CPLD_LED2_SHIFT_CTR4) & CPLD_LED2_MASK_CTR4)

#define CPLD_BID_MASK_REV        0x0f

#define CPLD_WDC_MASK_TIME       0x3f

#define CPLD_I2C_DBG0_STATE_MASK 0x1f
#define CPLD_I2C_DBG0_STATE_RDY  0x03

#define CPLD_I2C_C1_ID_READ      0x04
#define CPLD_I2C_C1_DIAG_WRITE   0x02
#define CPLD_I2C_C1_DIAG_READ    0x01

#define CPLD_I2C_S1_READY        0x01
#define CPLD_I2C_S1_ACCESS_BAD   0x02

#define CPLD_I2C_C5_RESET         (1<<0)

/* GPIR is active low */
#define CPLD_GPIR_BIT_WD_DIS_N     0x02
#define CPLD_GPIR_BIT_FLASH_PROG_N 0x04
#define CPLD_GPIR_BIT_TEST_N       0x10

#define CPLD_PROG_BIT_PROG       0x01

#define CPLD_INIT_BIT_INIT       0x01

#define CPLD_DONE_BIT_DONE       0x01

/* CPLD SW* registers are mapped into irqs */

#define CPLD_VER_MASK_MINOR      0x0f
#define CPLD_VER_MASK_MAJOR      0xf0
#define CPLD_VER_SHIFT_MAJOR     4

#define CPLD_SJMPR_BIT_WD_DIS    0x01
#define CPLD_SJMPR_BIT_WD_NMI    0x02

#define CPLD_JCSTAT_BITV         0x01

#define CPLD_RFCI2CWR_BIT_RF0_EEPROM 0x01
#define CPLD_RFCI2CWR_BIT_RF1_EEPROM 0x02
#define CPLD_RFCI2CWR_BIT_RF2_EEPROM 0x04

#define CPLD_CMDR_STOP        (1<<6)
#define CPLD_CMDR_RD       (1<<5)
#define CPLD_CMDR_NACK        (1<<3)
#define CPLD_CMDR_CKSDIS      (1<<2)


#ifdef __KERNEL__
/****************************************************************************
 * Macros for reading/writing CPLD registers
 ***************************************************************************/

static inline void cpld_lock(void)
{
   fzm_cpldspi_lock(FZM_CPLDSPI_CPLD);
}

static inline int cpld_lock_interruptible(void)
{
   return fzm_cpldspi_lock_interruptible(FZM_CPLDSPI_CPLD);
}

static inline int cpld_trylock(void)
{
   return fzm_cpldspi_trylock(FZM_CPLDSPI_CPLD);
}

static inline void cpld_unlock(void)
{
   fzm_cpldspi_unlock(FZM_CPLDSPI_CPLD);
}

static inline u8 readb_cpld_offset(u32 offset)
{
   u8 val = 0;
   fzm_cpldspi_read(FZM_CPLDSPI_CPLD,
                    offset, &val, sizeof(val));
   return val;
}

static inline int writeb_cpld_offset(u32 offset, u8 val)
{
   return fzm_cpldspi_write(FZM_CPLDSPI_CPLD,
                            offset, &val, sizeof(val));
}

static inline u8 readb_cpld_offset_locked(u32 offset)
{
   u8 val = 0;
   fzm_cpldspi_read_locked(FZM_CPLDSPI_CPLD,
                           offset, &val, sizeof(val));
   return val;
}

static inline int writeb_cpld_offset_locked(u32 offset, u8 val)
{
   return fzm_cpldspi_write_locked(FZM_CPLDSPI_CPLD,
                                   offset, &val, sizeof(val));
}

/* define all functions for reading */
#define REG(x) \
   static inline u8 readb_cpld_##x(void) \
   { \
      return readb_cpld_offset(offsetof(struct cpld_regs, x)); \
   }
#define REGW(x)
#define REGR(x) REG(x)
#define REGG(x) REG(x)
#define RESV(n)
#  include "fzm_cpld_regs.def"


/* define all functions for writing */
#define REG(x) \
   static inline int writeb_cpld_##x(u8 val) \
   { \
      return writeb_cpld_offset(offsetof(struct cpld_regs, x), val); \
   }
#define REGW(x) REG(x)
#define REGR(x)
#define REGG(x) REG(x)
#define RESV(n)
#  include "fzm_cpld_regs.def"


/* define all functions for locked reading */
#define REG(x) \
   static inline u8 readb_cpld_##x##_locked(void) \
   { \
      return readb_cpld_offset_locked(offsetof(struct cpld_regs, x)); \
   }
#define REGW(x)
#define REGR(x) REG(x)
#define REGG(x) REG(x)
#define RESV(n)
#  include "fzm_cpld_regs.def"


/* define all functions for locked writing */
#define REG(x) \
   static inline int writeb_cpld_##x##_locked(u8 val) \
   { \
      return writeb_cpld_offset_locked(offsetof(struct cpld_regs, x), val); \
   }
#define REGW(x) REG(x)
#define REGR(x)
#define REGG(x) REG(x)
#define RESV(n)
#  include "fzm_cpld_regs.def"

/* return the RAW IRQ status from swstat* for a virq */
extern int cpld_ic_raw_status(unsigned int virq);

#else
/* userspace */
#  include <unistd.h>
#  include <errno.h>

static inline int readb_cpld(int dev, unsigned int reg)
{
   int ret;
   __u8 val;
   ret = pread(dev, &val, sizeof(val), reg);
   if (-1 == ret)
      return -errno;
   return val;
}

static inline int writeb_cpld(int dev, unsigned int reg, __u8 val)
{
   int ret;
   ret = pwrite(dev, &val, sizeof(val), reg);
   if (-1 == ret)
      return -errno;
   return 0;
}

/* define all functions for reading */
#define REG(x) \
   static inline int readb_cpld_##x(int dev) \
   { \
      return readb_cpld(dev, \
                        offsetof(struct cpld_regs, x)); \
   }
#define REGW(x)
#define REGR(x) REG(x)
#define REGG(x) REG(x)
#define RESV(n)
#  include "fzm_cpld_regs.def"


/* define all functions for writing */
#define REG(x) \
   static inline int writeb_cpld_##x(int dev, __u8 val) \
   { \
      return writeb_cpld(dev, \
                         offsetof(struct cpld_regs, x), val); \
   }
#define REGW(x) REG(x)
#define REGR(x)
#define REGG(x) REG(x)
#define RESV(n)
#  include "fzm_cpld_regs.def"

#endif

#endif
