#include <linux/module.h>
#include <linux/keystone-dma.h>
#include <linux/version.h>
#include <net/sock.h>
#include <linux/interrupt.h>

#define BITS(x) (BIT(x) - 1)

// TODO: remove this header file and replace with the real Syscom header
#include "SicadInfo.h"

#define FZM_NID 1

#include "msg_rtr_arm.h"
#include "AaSysComNid.h"
#include "ARM_DSP_comm_shared.h"
#include "ddal_k2mr.h"

#define MAX_DSPS 8
#define ERR_FLOW 65535

#define TX_POLL_BUDGET 32

#define MSG_RTR_ARM_SGLIST_SIZE  (MAX_SKB_FRAGS + 2)
#define MSG_RTR_ARM_PSDATA_LEN        16
#define DESC_FLOWTAG_SHIFT      16

#define MSG_RTR_ARM_MAX_CALLBACKS 2
#define MSG_RTR_ARM_MAX_REGISTERED_NETLINK_NID 4
#define MSG_RTR_ARM_NETLINK_NID_MASK 0x3

#define MSG_RTR_ARM_NUM_RX_QUEUES 2

#define MSG_RTR_ARM_SICAD_WORD_OFFSET 2
#define MSG_RTR_ARM_SICAD_MASK 0x0700
#define MSG_RTR_ARM_SICAD_SHIFT 8

#define LEAST_STARVATION_QUEUE 736
#define GREATEST_STARVATION_QUEUE 799
#define QUEUE_CONFIG_BASE_REGISTER 0x2a02000

//#define SMALL_PACKETS_QUEUE_DEPTH 896;
//#define LARGE_PACKETS_QUEUE_DEPTH 128;
#define SMALL_PACKETS_QUEUE_DEPTH 448;
#define LARGE_PACKETS_QUEUE_DEPTH 64;

#define MSG_RTR_ARM_TX_OK 0

//#define MSG_RTR_ARM_INLINE_TX_RECYCLE

//#define MSG_RTR_ARM_KM_CALLTRACE
#ifdef MSG_RTR_ARM_KM_CALLTRACE
#define CALLTRACE(format,...) printk(KERN_CRIT "[MRA] %s:%u called: "format"\n", __func__, __LINE__, ##__VA_ARGS__);
#else
#define CALLTRACE(format,...)
#endif

// #define MSG_RTR_ARM_KM_DEBUG
#ifdef MSG_RTR_ARM_KM_DEBUG
//#define DBG(format,...) printk(KERN_DEBUG "[MRA] %s:%u "format"\n", __func__, __LINE__, ##__VA_ARGS__);
#define DBG(format,...) printk(KERN_CRIT "[MRA] %s:%u "format"\n", __func__, __LINE__, ##__VA_ARGS__);
#else
#define DBG(format,...)
#endif


#define ERR(format,...)

/* #define ERR(format,...) printk(KERN_CRIT "[MRA] %s:%u "format"\n", __func__, __LINE__, ##__VA_ARGS__); */

#define RX_DEVICE (&rx_struct->rx_channel->dev->device)
#define TX_DEVICE (&tx_struct->tx_channel->dev->device)

static unsigned rx_poll_budget = 320;
module_param(rx_poll_budget, uint, 0644);
MODULE_PARM_DESC(rx_poll_budget, "rx packet processing budget in each tasklet");

static struct workqueue_struct *msg_rtr_arm_wq;

struct msg_rtr_arm_callback_mapping {
	u32 sicad;
	msgRtrARMCb cb;
};

struct msg_rtr_arm_dsp_flow {
	u32 core;
	u32 flow;
};

struct msg_rtr_arm_netlink_nid {
	u32 nid;
	u32 netlink_pid;
};

// Assume last 4 digits as core id to speed up lookup 
static int core2flowmap[] = {-1,
     ADCOMM_RX_FLOW_ID_DSP0_RX,
     ADCOMM_RX_FLOW_ID_DSP1_RX,
     ADCOMM_RX_FLOW_ID_DSP2_RX,
     ADCOMM_RX_FLOW_ID_DSP3_RX,
     ADCOMM_RX_FLOW_ID_DSP4_RX,
     ADCOMM_RX_FLOW_ID_DSP5_RX,
     ADCOMM_RX_FLOW_ID_DSP6_RX,
     ADCOMM_RX_FLOW_ID_DSP7_RX,
     14,
     14,
     -1,
     -1,
     -1,
     -1,
     -1
};
     

#if 0
static struct msg_rtr_arm_dsp_flow msg_rtr_arm_flow_map[] = {
	{NIDDSPCore1, ADCOMM_RX_FLOW_ID_DSP0_RX},
	{NIDDSPCore2, ADCOMM_RX_FLOW_ID_DSP1_RX},
	{NIDDSPCore3, ADCOMM_RX_FLOW_ID_DSP2_RX},
	{NIDDSPCore4, ADCOMM_RX_FLOW_ID_DSP3_RX},
	{NIDDSPCore5, ADCOMM_RX_FLOW_ID_DSP4_RX},
	{NIDDSPCore6, ADCOMM_RX_FLOW_ID_DSP5_RX},
	{NIDDSPCore7, ADCOMM_RX_FLOW_ID_DSP6_RX},
	{NIDDSPCore8, ADCOMM_RX_FLOW_ID_DSP7_RX},
        {0x9, 14},  // CCS_RT
        {0xA, 14}   // CCS_RT realtime
};
#endif

#define MSG_RTR_ARM_STAT_INC(field) msg_rtr_arm_global_data.msg_rtr_arm->stats.field++;
#define MSG_RTR_ARM_STAT_ADD(field, val) msg_rtr_arm_global_data.msg_rtr_arm->stats.field += (val);

struct msg_rtr_arm_stats {
	u32 rx_dsp_pkt;
	u32 rx_nl_pkt;
	u32 rx_nl_ctrl_pkt;
	u32 rx_vtc_pkt;
	u32 tx_dsp_pkt;
	u32 tx_nl_pkt;
	u32 tx_vtc_pkt;
	u32 rx_int;
	u32 rx_q0_alloc;
	u32 rx_q1_alloc;
	u32 rx_q0_alloc_fail;
	u32 rx_q1_alloc_fail;
	u32 rx_dma_fail;
	u32 tx_dma_fail;
	u32 tx_nl_fail;
	u32 tx_dropped_bad_dest;
	u32 tx_dropped_map_fail;
	u32 tx_dropped_prep_fail;
	u32 tx_dropped_submit_fail;
	u32 tx_dropped_internal;
	u32 tx_dropped_total;
	u32 tx_trace_pkt;
	u32 rx_trace_pkt;
	u32 rx_q0_starvation;
	u32 rx_q1_starvation;

	u32 rx2_q0_alloc;
	u32 rx2_q1_alloc;
	u32 rx2_q0_alloc_fail;
	u32 rx2_q1_alloc_fail;
	u32 rx2_dma_fail;
	u32 rx2_q0_starvation;
	u32 rx2_q1_starvation;
};

struct starvation_reg {
	u32 r;
};

struct msg_rtr_arm_tx_struct {
	const char* tx_chan_name;
	struct dma_chan* tx_channel;
	u32 tx_queue_depth;
	atomic_t tx_poll_count;
	dma_async_tx_callback callback;
#ifndef MSG_RTR_ARM_INLINE_TX_RECYCLE
	struct tasklet_struct	tx_poll_tasklet;
#endif
};

struct msg_rtr_arm_rx_struct {
    	const char * rx_chan_name;
	struct dma_chan *rx_channel;
	u32 rx_queue_depths[MSG_RTR_ARM_NUM_RX_QUEUES];
	u32 rx_buffer_sizes[MSG_RTR_ARM_NUM_RX_QUEUES];
	u32 rx_reserve_headroom;
	u32 rx_free_q0;
	u32 rx_free_q1;
	struct starvation_reg __iomem *rx_free_reg_0;
	struct tasklet_struct	rx_poll_tasklet;
};

struct msg_rtr_arm_priv {
	int dev_id;

	struct msg_rtr_arm_rx_struct rx_vtc;
	struct msg_rtr_arm_rx_struct rx_cp;

	struct msg_rtr_arm_tx_struct tx_vtc;
	struct msg_rtr_arm_tx_struct tx_cp;
	u32 dma_flow_loopback;
	struct sock *nl_sock;
	u32 nl_linux_pid;
	struct msg_rtr_arm_callback_mapping cbmap[MSG_RTR_ARM_MAX_CALLBACKS];
	struct msg_rtr_arm_stats stats;
	struct proc_dir_entry * proc_file;
	struct proc_dir_entry * proc_file_raw;
	bool loopback;
	bool rx_trace;
	bool tx_trace;
	struct msg_rtr_arm_netlink_nid netlink_nid[MSG_RTR_ARM_MAX_REGISTERED_NETLINK_NID];
};

struct msg_rtr_arm_global {
	struct msg_rtr_arm_priv *msg_rtr_arm;
};

static struct msg_rtr_arm_global msg_rtr_arm_global_data;

struct msg_rtr_arm_packet {
	void                            *primary_bufptr;
	unsigned int                     primary_bufsiz;
	int                              sg_ents;
	dma_cookie_t                     cookie;
	struct scatterlist               sg[MSG_RTR_ARM_SGLIST_SIZE];
};

static struct kmem_cache* pinfo_cache;

// Params for lookback test
static int tst_loopback = 0;
static uint tst_loopback_flow = 11; 
module_param(tst_loopback, int, 0644);
module_param(tst_loopback_flow, uint, 0644);

int MsgRtrARMRegisterSysComID(u16 sysComID, msgRtrARMCb callback)
{
	u32 i;

	CALLTRACE();

	for (i = 0; i < MSG_RTR_ARM_MAX_CALLBACKS; i++)
	{
		if (msg_rtr_arm_global_data.msg_rtr_arm->cbmap[i].sicad == 0 ||
			msg_rtr_arm_global_data.msg_rtr_arm->cbmap[i].sicad == sysComID)
		{
			DBG("sicad:%x", sysComID);
			msg_rtr_arm_global_data.msg_rtr_arm->cbmap[i].sicad = sysComID;
			msg_rtr_arm_global_data.msg_rtr_arm->cbmap[i].cb = callback;
			return 0;
		}
	}
	return -1;
}
EXPORT_SYMBOL(MsgRtrARMRegisterSysComID);

int MsgRtrARMUnregister(u16 sysComID)
{
	u32 i;

	CALLTRACE();

	for (i = 0; i < MSG_RTR_ARM_MAX_CALLBACKS; i++)
	{
		if (msg_rtr_arm_global_data.msg_rtr_arm->cbmap[i].sicad == sysComID)
		{
			msg_rtr_arm_global_data.msg_rtr_arm->cbmap[i].sicad = 0;
			msg_rtr_arm_global_data.msg_rtr_arm->cbmap[i].cb = NULL;
			return 0;
		}
	}
	return -1;
}
EXPORT_SYMBOL(MsgRtrARMUnregister);


static void msg_rtr_arm_register_netlink_pid(u32 nid, u32 netlink_pid)
{
	u32 i = nid & MSG_RTR_ARM_NETLINK_NID_MASK;

	CALLTRACE();

	if (msg_rtr_arm_global_data.msg_rtr_arm->netlink_nid[i].nid == 0 || msg_rtr_arm_global_data.msg_rtr_arm->netlink_nid[i].nid == nid)
	{
		msg_rtr_arm_global_data.msg_rtr_arm->netlink_nid[i].nid = nid;
		msg_rtr_arm_global_data.msg_rtr_arm->netlink_nid[i].netlink_pid = netlink_pid;
		return;
	}
	printk(KERN_CRIT "MSG_RTR_ARM: registration conflict! %x, %u, previous nid: %x\n", nid, netlink_pid, msg_rtr_arm_global_data.msg_rtr_arm->netlink_nid[i].nid);
}

static void msg_rtr_arm_tx_complete_cp(void *data);
static void msg_rtr_arm_tx_complete_vtc(void *data);
#ifndef MSG_RTR_ARM_INLINE_TX_RECYCLE
static void msg_rtr_arm_tx_notify(struct dma_chan *chan, void *arg);
#endif

static int msg_rtr_arm_open_tx_one(struct msg_rtr_arm_tx_struct* tx)
{
	struct dma_keystone_info config;
	dma_cap_mask_t mask;
	int ret;

	CALLTRACE();

	dma_cap_zero(mask);
	dma_cap_set(DMA_SLAVE, mask);

	tx->tx_channel = dma_request_channel_by_name(mask, tx->tx_chan_name);
	if (IS_ERR_OR_NULL(tx->tx_channel)) {
		ERR("Could not get DMA channel \"%s\" (%u)",
			tx->tx_chan_name, (unsigned int)tx->tx_channel);
		tx->tx_channel = NULL;
		return -ENODEV;
	}

	memset(&config, 0, sizeof(config));
	config.direction = DMA_MEM_TO_DEV;
	config.tx_queue_depth = tx->tx_queue_depth;
	ret = dma_keystone_config(tx->tx_channel, &config);
	if (ret) {
		dev_err(&tx->tx_channel->dev->device,
			"Could not configure DMA channel \"%s\": %d\n",
			tx->tx_chan_name, ret);
		dma_release_channel(tx->tx_channel);
		tx->tx_channel = NULL;
		return ret;
	}

	atomic_set(&tx->tx_poll_count, tx->tx_queue_depth);

	dev_dbg(&tx->tx_channel->dev->device, "opened tx pipe %s\n",
		tx->tx_chan_name);

	return 0;
}

static int msg_rtr_arm_open_tx(struct msg_rtr_arm_priv *msg_rtr_arm)
{
	int err = 0;

	CALLTRACE();

	err = msg_rtr_arm_open_tx_one(&msg_rtr_arm->tx_vtc);
	if (err)
		return err;

	err = msg_rtr_arm_open_tx_one(&msg_rtr_arm->tx_cp);
	if (err)
		return err;

	msg_rtr_arm->tx_cp.callback = msg_rtr_arm_tx_complete_cp;
	msg_rtr_arm->tx_vtc.callback = msg_rtr_arm_tx_complete_vtc;

#ifndef MSG_RTR_ARM_INLINE_TX_RECYCLE
	dma_set_notify(msg_rtr_arm->tx_vtc.tx_channel, msg_rtr_arm_tx_notify, &msg_rtr_arm->tx_vtc);
	dma_set_notify(msg_rtr_arm->tx_cp.tx_channel, msg_rtr_arm_tx_notify, &msg_rtr_arm->tx_cp);
#endif

	return 0;
}

static void msg_rtr_arm_close_channel(struct dma_chan* chan)
{

	CALLTRACE();

	if (!chan)
		return;

	dmaengine_pause(chan);
	dma_poll(chan, -1);
	dma_release_channel(chan);
}

/* Close the device */
static int msg_rtr_arm_close(struct msg_rtr_arm_priv *msg_rtr_arm)
{
	u32 reg_ptr;

	CALLTRACE();

	msg_rtr_arm_close_channel(msg_rtr_arm->rx_vtc.rx_channel);
	msg_rtr_arm->rx_vtc.rx_channel = NULL;
        msg_rtr_arm_close_channel(msg_rtr_arm->rx_cp.rx_channel);
        msg_rtr_arm->rx_cp.rx_channel = NULL;

	msg_rtr_arm_close_channel(msg_rtr_arm->tx_vtc.tx_channel);
	msg_rtr_arm->tx_vtc.tx_channel = NULL;

        if(msg_rtr_arm->tx_cp.tx_channel != NULL)
	  msg_rtr_arm_close_channel(msg_rtr_arm->tx_cp.tx_channel);
	msg_rtr_arm->tx_cp.tx_channel = NULL;

	if (msg_rtr_arm->nl_sock)
	{
		netlink_kernel_release(msg_rtr_arm->nl_sock);
	}

	flush_workqueue( msg_rtr_arm_wq );
	destroy_workqueue( msg_rtr_arm_wq );

	if (msg_rtr_arm->proc_file)
	{
		remove_proc_entry("msg_rtr_arm", NULL);
	}
	if (msg_rtr_arm->proc_file_raw)
	{
		remove_proc_entry("msg_rtr_arm_raw", NULL);
	}

	if (msg_rtr_arm->rx_vtc.rx_free_reg_0)
	{
		reg_ptr = QUEUE_CONFIG_BASE_REGISTER + 0x20 + ((msg_rtr_arm->rx_vtc.rx_free_q0 - LEAST_STARVATION_QUEUE) >> 2) * 4;
		release_mem_region(reg_ptr,1);
	}
	if (msg_rtr_arm->rx_cp.rx_free_reg_0)
	{
		reg_ptr = QUEUE_CONFIG_BASE_REGISTER + 0x20 + ((msg_rtr_arm->rx_cp.rx_free_q0 - LEAST_STARVATION_QUEUE) >> 2) * 4;
		release_mem_region(reg_ptr,1);
	}

	kfree(msg_rtr_arm);

	kmem_cache_destroy(pinfo_cache);
	pinfo_cache = NULL;

	tasklet_kill(&msg_rtr_arm->rx_vtc.rx_poll_tasklet);
        tasklet_kill(&msg_rtr_arm->rx_cp.rx_poll_tasklet);
#ifndef MSG_RTR_ARM_INLINE_TX_RECYCLE
	tasklet_kill(&msg_rtr_arm->tx_vtc.tx_poll_tasklet);
	tasklet_kill(&msg_rtr_arm->tx_cp.tx_poll_tasklet);
#endif

	return 0;
}

static int msg_rtr_arm_send_netlink_msg(struct msg_rtr_arm_priv * msg_rtr_arm, struct sk_buff * skb, u32 nl_pid, bool trace)
{
	struct nlmsghdr * nlh;
	ddal_k2mr_msg_rtr_arm_header_t * k2mr_hdr;

	CALLTRACE();

	k2mr_hdr = (ddal_k2mr_msg_rtr_arm_header_t *)skb_push(skb, sizeof(ddal_k2mr_msg_rtr_arm_header_t));
	k2mr_hdr->type = trace ? DDAL_K2MR_TRACE_PACKET : DDAL_IGNORE_CONTROL_TYPE;

	nlh = (struct nlmsghdr *)skb_push(skb, NLMSG_HDRLEN); //We should have enough reserved, but probably should check for failure
	nlh->nlmsg_len = skb->len;
	nlh->nlmsg_type = 0;
	nlh->nlmsg_flags = 0;
	nlh->nlmsg_seq = 0; //Is there any issue with 0 sequence numbers for everything?
	nlh->nlmsg_pid = 0; //doublecheck that this should be 0, and no the nl_linux_pid

	DBG("Attempting to send netlink message of size: %u, pid: %u\n", skb->len, msg_rtr_arm->nl_linux_pid);

	return(netlink_unicast(msg_rtr_arm->nl_sock, skb, nl_pid, MSG_DONTWAIT));
	//netlink_unicast frees on failure, no action is needed
}

static int msg_rtr_arm_netlink_classify_and_send(struct sk_buff *skb, u32 nid)
{
	struct msg_rtr_arm_priv *msg_rtr_arm = msg_rtr_arm_global_data.msg_rtr_arm;
	u32 i = nid & MSG_RTR_ARM_NETLINK_NID_MASK;
	struct SMsgData *sysComBuf;
	TAaSysComNid target;
	TAaSysComNid sender;
	TAaSysComMsgId msgId;
        struct msg_rtr_arm_rx_struct* rx_struct = &msg_rtr_arm->rx_cp;

	CALLTRACE();

	if (msg_rtr_arm->netlink_nid[i].nid == nid)
	{
		if(WARN_ON(msg_rtr_arm->netlink_nid[i].netlink_pid == 0))
		{
			sysComBuf = (struct SMsgData*)skb->data;
			target = (TAaSysComNid)(sysComBuf->header.target);
			sender = (TAaSysComNid)(sysComBuf->header.sender);
			msgId  = (TAaSysComMsgId)(sysComBuf->header.msgId);
			dev_err(RX_DEVICE, "Error sending skb through netlink, pid is 0. sender 0x%X target 0x%X msgId %u\n",
						sender, target, msgId);
			dev_kfree_skb(skb);
			return -EDESTADDRREQ;
		}

		//TODO: Make this a unique stat
		MSG_RTR_ARM_STAT_INC(tx_nl_pkt);
		if (msg_rtr_arm_send_netlink_msg(msg_rtr_arm, skb, msg_rtr_arm->netlink_nid[i].netlink_pid, false) < 0)
		{
			MSG_RTR_ARM_STAT_INC(tx_nl_fail); //TODO: change to unique stat
			dev_warn_ratelimited(RX_DEVICE, "msg_rtr_arm_send_netlink_msg failed to %x\n", nid);
			//netlink_unicast frees skb on failure, no action is needed here
		}
		return 0;
	}

	//Else, send to linux using netlink
	if (msg_rtr_arm->nl_linux_pid != 0)
	{
		MSG_RTR_ARM_STAT_INC(tx_nl_pkt);
		if (msg_rtr_arm_send_netlink_msg(msg_rtr_arm, skb, msg_rtr_arm->nl_linux_pid, false) < 0)
		{
			MSG_RTR_ARM_STAT_INC(tx_nl_fail);
			dev_warn_ratelimited(RX_DEVICE, "msg_rtr_arm_send_netlink_msg failed\n");
			//netlink_unicast frees skb on failure, no action is needed here
		}
		return 0;
	}
	else
	{
		dev_kfree_skb_any(skb);
		return -ENXIO;
	}
}

static inline void msg_rtr_arm_frag_free(bool is_frag, void *ptr)
{

	CALLTRACE();

	if (is_frag)
		put_page(virt_to_head_page(ptr));
	else
		kfree(ptr);
}

static void msg_rtr_arm_rx_complete(void *data, struct msg_rtr_arm_rx_struct* rx_struct);

static void msg_rtr_arm_rx_complete_cp(void *data) {
    MSG_RTR_ARM_STAT_INC(rx_dsp_pkt);
    msg_rtr_arm_rx_complete(data, &msg_rtr_arm_global_data.msg_rtr_arm->rx_cp);
}

static void msg_rtr_arm_rx_complete_vtc(void *data) {
    msg_rtr_arm_rx_complete(data, &msg_rtr_arm_global_data.msg_rtr_arm->rx_vtc);
}

static void msg_rtr_arm_rx_complete(void *data, struct msg_rtr_arm_rx_struct* rx_struct)
{
	struct msg_rtr_arm_packet *p_info = data;
	struct msg_rtr_arm_priv *msg_rtr_arm = msg_rtr_arm_global_data.msg_rtr_arm;
	struct sk_buff *skb;
	struct scatterlist *sg;
	enum dma_status status;
	unsigned int frags;
	unsigned int len;

	struct SMsgData *sysComBuf;
	TAaSysComNid sicad;
	u32 i;

	CALLTRACE();

	status = dma_async_is_tx_complete(rx_struct->rx_channel,
		p_info->cookie, NULL, NULL);
	WARN_ON(status != DMA_SUCCESS && status != DMA_ERROR);

	/* sg[0] describes the primary buffer */
	/* Build a new sk_buff for this buffer */
	dma_unmap_single(RX_DEVICE, sg_dma_address(&p_info->sg[0]),
		p_info->primary_bufsiz, DMA_FROM_DEVICE);
	DBG("Building the skb: buf:%p len:%d", p_info->primary_bufptr, p_info->primary_bufsiz);
	skb = build_skb(p_info->primary_bufptr, p_info->primary_bufsiz);
	if (unlikely(!skb)) {
		/* Free the primary buffer */
		msg_rtr_arm_frag_free((p_info->primary_bufsiz <= PAGE_SIZE),
			p_info->primary_bufptr);
		/* Free other buffers in the scatterlist */
		for (frags = 0, sg = sg_next(&p_info->sg[0]);
			frags < MSG_RTR_ARM_SGLIST_SIZE-1 && sg;
				 ++frags, sg = sg_next(sg)) {
			dma_unmap_page(RX_DEVICE, sg_dma_address(sg),
				PAGE_SIZE, DMA_FROM_DEVICE);
			__free_page(sg_page(sg));
		}
		return;
	}

	/* Update data, tail, and len */
	skb_reserve(skb, rx_struct->rx_reserve_headroom);
	len = sg_dma_len(&p_info->sg[0]);
	__skb_put(skb, len);

	/* Fill in the page fragment list from sg[1] and later */
	for (frags = 0, sg = sg_next(&p_info->sg[0]);
		frags < MSG_RTR_ARM_SGLIST_SIZE-1 && sg;
			 ++frags, sg = sg_next(sg)) {
		dma_unmap_page(RX_DEVICE, sg_dma_address(sg),
			PAGE_SIZE, DMA_FROM_DEVICE);
		len = sg_dma_len(sg);
		skb_add_rx_frag(skb, frags, sg_page(sg), sg->offset, len, PAGE_SIZE);
	}

	if (unlikely(status != DMA_SUCCESS)) {
		dev_warn(RX_DEVICE,
			"receive: reclaimed packet %p, status %d\n",
			p_info, status);
		dev_kfree_skb(skb);
		kmem_cache_free(pinfo_cache, p_info);
		MSG_RTR_ARM_STAT_INC(rx_dma_fail);
		return;
	}

	if (unlikely(!skb->len)) {
		dev_warn(RX_DEVICE, "receive: zero length packet\n");
		dev_kfree_skb(skb);
		kmem_cache_free(pinfo_cache, p_info);
		MSG_RTR_ARM_STAT_INC(rx_dma_fail);
		return;
	}

	kmem_cache_free(pinfo_cache, p_info);

	//Extract the sicad from the received buffer
	sysComBuf = (struct SMsgData*)skb->data;
	sicad = (TAaSysComNid)(sysComBuf->header.target >> 16);

	DBG("sicad:%x sysComBuf:%p", sicad, sysComBuf);

	if (msg_rtr_arm_global_data.msg_rtr_arm->rx_trace) {
		MSG_RTR_ARM_STAT_INC(rx_trace_pkt);
		msg_rtr_arm_send_netlink_msg(msg_rtr_arm_global_data.msg_rtr_arm,skb_clone(skb, GFP_ATOMIC),msg_rtr_arm->nl_linux_pid, true);
	}

	//Determine if this sicad has a registered callback
	for (i = 0; i < MSG_RTR_ARM_MAX_CALLBACKS; i++)
	{
                #if 0
                printk(KERN_CRIT "[MRA] ux:%x:%x\n",i, sicad, msg_rtr_arm_global_data.msg_rtr_arm->cbmap[i].sicad);
                #endif
		if (msg_rtr_arm_global_data.msg_rtr_arm->cbmap[i].sicad == sicad &&
			msg_rtr_arm_global_data.msg_rtr_arm->cbmap[i].cb)
		{
			DBG("Attempting to call registered callback (sicad:%04x buf:%p cb:%p skb:%p)", sicad, sysComBuf, (*(msg_rtr_arm_global_data.msg_rtr_arm->cbmap[i].cb)), skb);
			MSG_RTR_ARM_STAT_INC(tx_vtc_pkt);
			(*(msg_rtr_arm_global_data.msg_rtr_arm->cbmap[i].cb))(skb);
			return;
		}
	}

	msg_rtr_arm_netlink_classify_and_send(skb, (u32)sicad);
}

static inline void msg_rtr_arm_update_starvation_counters(struct msg_rtr_arm_priv *msg_rtr_arm)
{
	u32 tmp, orig_tmp;
	u32 rx_free_index_0 = msg_rtr_arm->rx_cp.rx_free_q0;
	u32 rx_free_index_1 = msg_rtr_arm->rx_cp.rx_free_q1;

	CALLTRACE();

	if (msg_rtr_arm->rx_cp.rx_free_reg_0)
	{
		orig_tmp = __raw_readl(&msg_rtr_arm->rx_cp.rx_free_reg_0->r);
		if (orig_tmp != 0)
		{
			DBG("cp starvation register non-zero: %x\n", orig_tmp);
		}
		tmp = ((orig_tmp >> ((rx_free_index_0 % 4) * 8)) & 0xFF);
		msg_rtr_arm->stats.rx_q0_starvation += tmp;

		if (msg_rtr_arm->rx_cp.rx_free_q1 >= LEAST_STARVATION_QUEUE &&
			msg_rtr_arm->rx_cp.rx_free_q1 <= GREATEST_STARVATION_QUEUE)
		{
			tmp = ((orig_tmp >> ((rx_free_index_1 % 4) * 8)) & 0xFF);
			msg_rtr_arm->stats.rx_q1_starvation += tmp;
		}
	}
}

static inline void msg_rtr_arm_update_starvation_counters2(struct msg_rtr_arm_priv *msg_rtr_arm)
{
	u32 tmp, orig_tmp;
	u32 rx_free_index_0 = msg_rtr_arm->rx_vtc.rx_free_q0;
	u32 rx_free_index_1 = msg_rtr_arm->rx_vtc.rx_free_q1;

	CALLTRACE();

	if (msg_rtr_arm->rx_vtc.rx_free_reg_0)
	{
		orig_tmp = __raw_readl(&msg_rtr_arm->rx_vtc.rx_free_reg_0->r);
		if (orig_tmp != 0)
		{
			DBG("vtc starvation register non-zero: %x\n", orig_tmp);
		}
		tmp = ((orig_tmp >> ((rx_free_index_0 % 4) * 8)) & 0xFF);
		msg_rtr_arm->stats.rx2_q0_starvation += tmp;

		if (msg_rtr_arm->rx_vtc.rx_free_q1 >= LEAST_STARVATION_QUEUE &&
			msg_rtr_arm->rx_vtc.rx_free_q1 <= GREATEST_STARVATION_QUEUE)
		{
			tmp = ((orig_tmp >> ((rx_free_index_1 % 4) * 8)) & 0xFF);
			msg_rtr_arm->stats.rx2_q1_starvation += tmp;
		}
	}
}


static void msg_rtr_arm_rx_poll(unsigned long arg)
{
	u32 packets = rx_poll_budget;
	struct msg_rtr_arm_priv *msg_rtr_arm = msg_rtr_arm_global_data.msg_rtr_arm;

	CALLTRACE();

	packets = dma_poll(msg_rtr_arm->rx_cp.rx_channel, rx_poll_budget);
	dma_rxfree_refill(msg_rtr_arm->rx_cp.rx_channel);

	if (packets >= rx_poll_budget)
	{
		tasklet_schedule(&msg_rtr_arm->rx_cp.rx_poll_tasklet);
	}
	else
	{
		dmaengine_resume(msg_rtr_arm->rx_cp.rx_channel);
	}

	msg_rtr_arm_update_starvation_counters(msg_rtr_arm);
}

static void msg_rtr_arm_rx_poll2(unsigned long arg)
{
	u32 packets = rx_poll_budget;
	struct msg_rtr_arm_priv *msg_rtr_arm = msg_rtr_arm_global_data.msg_rtr_arm;

	CALLTRACE();

	packets = dma_poll(msg_rtr_arm->rx_vtc.rx_channel, rx_poll_budget);
	dma_rxfree_refill(msg_rtr_arm->rx_vtc.rx_channel);

	if (packets >= rx_poll_budget)
	{
		tasklet_schedule(&msg_rtr_arm->rx_vtc.rx_poll_tasklet);
	}
	else
	{
		dmaengine_resume(msg_rtr_arm->rx_vtc.rx_channel);
	}

	msg_rtr_arm_update_starvation_counters2(msg_rtr_arm);
}

#ifndef MSG_RTR_ARM_INLINE_TX_RECYCLE
static void msg_rtr_arm_tx_poll(unsigned long arg)
{
	u32 packets = TX_POLL_BUDGET;
	struct msg_rtr_arm_tx_struct* tx = (struct msg_rtr_arm_tx_struct*)arg;

	CALLTRACE();

	while (packets >= TX_POLL_BUDGET)
	{
		packets = dma_poll(tx->tx_channel, TX_POLL_BUDGET);
	}
	dmaengine_resume(tx->tx_channel);
}

static void msg_rtr_arm_tx_notify(struct dma_chan *chan, void *arg)
{
	struct msg_rtr_arm_tx_struct * tx = arg;

	CALLTRACE();

	dmaengine_pause(tx->tx_channel);
	tasklet_schedule(&tx->tx_poll_tasklet);
}
#endif

static void msg_rtr_arm_rx_notify(struct dma_chan *chan, void *arg)
{
	struct msg_rtr_arm_rx_struct * rx = arg;

	CALLTRACE();

	dmaengine_pause(rx->rx_channel);
	tasklet_schedule(&rx->rx_poll_tasklet);
}

static void msg_rtr_arm_rx_complete2nd(void *data)
{
	struct page *page = data;

	CALLTRACE();

	WARN(1, "Attempt to complete secondary receive buffer!\n");
	/* FIXME: We need to unmap the page, but don't have the info needed */
	__free_page(page);
}

static struct dma_async_tx_descriptor *msg_rtr_arm_rxpool_alloc(void *arg,
	unsigned q_num, unsigned size)
{
	struct msg_rtr_arm_priv *msg_rtr_arm = arg;
	struct dma_async_tx_descriptor *desc;
	int err = 0;
        struct msg_rtr_arm_rx_struct* rx_struct = &msg_rtr_arm->rx_cp;

	CALLTRACE();

	if (q_num == 0) {
		struct msg_rtr_arm_packet *p_info;
		void *bufptr;
		unsigned char *data;
		unsigned int bufsiz;

		// DBG("msg_rtr_arm Attempting to alloc in pool 0");

		/* Allocate a primary receive queue entry */
		p_info = kmem_cache_alloc(pinfo_cache, GFP_ATOMIC);
		if (!p_info) {
			dev_err(RX_DEVICE, "packet alloc failed\n");
			MSG_RTR_ARM_STAT_INC(rx_q0_alloc_fail);
			return NULL;
		}

		bufsiz = SKB_DATA_ALIGN(size + msg_rtr_arm->rx_cp.rx_reserve_headroom) +
			SKB_DATA_ALIGN(sizeof(struct skb_shared_info));
		if (bufsiz <= PAGE_SIZE) {
			bufptr = netdev_alloc_frag(bufsiz);
			p_info->primary_bufsiz = bufsiz;
		} else {
			bufptr = kmalloc(bufsiz, GFP_ATOMIC | __GFP_COLD);
			p_info->primary_bufsiz = 0;
		}
		if (!bufptr) {
			dev_warn(RX_DEVICE, "Primary RX buffer allocation failed\n");
			kmem_cache_free(pinfo_cache, p_info);
			MSG_RTR_ARM_STAT_INC(rx_q0_alloc_fail);
			return NULL;
		}
		p_info->primary_bufptr = bufptr;

		/* Same as skb_reserve(skb, msg_rtr_arm->rx_reserve_headroom) */
		data = bufptr + msg_rtr_arm->rx_cp.rx_reserve_headroom;

		sg_init_table(p_info->sg, MSG_RTR_ARM_SGLIST_SIZE);
		sg_set_buf(&p_info->sg[0], data, size);

		p_info->sg_ents = dma_map_sg(RX_DEVICE, &p_info->sg[0],
			1, DMA_FROM_DEVICE);
		if (p_info->sg_ents != 1) {
			dev_err(RX_DEVICE, "dma map failed\n");
			msg_rtr_arm_frag_free((bufsiz <= PAGE_SIZE), bufptr);
			kmem_cache_free(pinfo_cache, p_info);
			MSG_RTR_ARM_STAT_INC(rx_q0_alloc_fail);
			return NULL;
		}

		desc = dmaengine_prep_slave_sg(msg_rtr_arm->rx_cp.rx_channel, p_info->sg,
			1, DMA_DEV_TO_MEM, 0);
		if (IS_ERR_OR_NULL(desc)) {
			dma_unmap_sg(RX_DEVICE, &p_info->sg[0], 1, DMA_FROM_DEVICE);
			msg_rtr_arm_frag_free((bufsiz <= PAGE_SIZE), bufptr);
			kmem_cache_free(pinfo_cache, p_info);
			err = PTR_ERR(desc);
			if (err != -ENOMEM) {
				dev_err(RX_DEVICE,
					"dma prep failed, error %d\n", err);
			}
			MSG_RTR_ARM_STAT_INC(rx_q0_alloc_fail);
			return NULL;
		}

		desc->callback_param = p_info;
		desc->callback = msg_rtr_arm_rx_complete_cp;
		p_info->cookie = desc->cookie;

		MSG_RTR_ARM_STAT_INC(rx_q0_alloc);

	} else {
		struct page *page;

		/* Allocate a secondary receive queue entry */
		struct scatterlist sg[1];

		// DBG("msg_rtr_arm Attempting to alloc in pool 1");

		page = alloc_page(GFP_ATOMIC | GFP_DMA32 | __GFP_COLD);
		if (!page) {
			dev_warn(RX_DEVICE, "page alloc failed for pool %d\n", q_num);
			MSG_RTR_ARM_STAT_INC(rx_q1_alloc_fail);
			return NULL;
		}

		sg_init_table(sg, 1);
		sg_set_page(&sg[0], page, PAGE_SIZE, 0);

		err = dma_map_sg(RX_DEVICE, sg, 1, DMA_FROM_DEVICE);
		if (err != 1) {
			dev_warn(RX_DEVICE, "map error for pool %d\n", q_num);
			__free_page(page);
			MSG_RTR_ARM_STAT_INC(rx_q1_alloc_fail);
			return NULL;
		}

		desc = dmaengine_prep_slave_sg(msg_rtr_arm->rx_cp.rx_channel, sg, 1,
			DMA_DEV_TO_MEM,
			q_num << DMA_QNUM_SHIFT);
		if (IS_ERR_OR_NULL(desc)) {
			dma_unmap_sg(RX_DEVICE, sg, 1, DMA_FROM_DEVICE);
			__free_page(page);

			err = PTR_ERR(desc);
			if (err != -ENOMEM) {
				dev_err(RX_DEVICE,
					"dma prep failed, error %d\n", err);
			}
			MSG_RTR_ARM_STAT_INC(rx_q1_alloc_fail);
			return NULL;
		}

		desc->callback_param = page;
		desc->callback = msg_rtr_arm_rx_complete2nd;

		MSG_RTR_ARM_STAT_INC(rx_q1_alloc);
	}

	wmb();
	return desc;
}

static struct dma_async_tx_descriptor *msg_rtr_arm_rxpool_alloc2(void *arg,
	unsigned q_num, unsigned size)
{
	struct msg_rtr_arm_priv *msg_rtr_arm = arg;
	struct dma_async_tx_descriptor *desc;
	int err = 0;
        struct msg_rtr_arm_rx_struct* rx_struct = &msg_rtr_arm->rx_vtc;

	CALLTRACE();

	if (q_num == 0) {
		struct msg_rtr_arm_packet *p_info;
		void *bufptr;
		unsigned char *data;
		unsigned int bufsiz;

		// DBG("vtc: msg_rtr_arm Attempting to alloc in pool 0");

		/* Allocate a primary receive queue entry */
		p_info = kmem_cache_alloc(pinfo_cache, GFP_ATOMIC);
		if (!p_info) {
			dev_err(RX_DEVICE, "vtc: packet alloc failed\n");
			MSG_RTR_ARM_STAT_INC(rx2_q0_alloc_fail);
			return NULL;
		}

		bufsiz = SKB_DATA_ALIGN(size + msg_rtr_arm->rx_vtc.rx_reserve_headroom) +
			SKB_DATA_ALIGN(sizeof(struct skb_shared_info));
		if (bufsiz <= PAGE_SIZE) {
			bufptr = netdev_alloc_frag(bufsiz);
			p_info->primary_bufsiz = bufsiz;
		} else {
			bufptr = kmalloc(bufsiz, GFP_ATOMIC | __GFP_COLD);
			p_info->primary_bufsiz = 0;
		}
		if (!bufptr) {
			dev_warn(RX_DEVICE, "vtc: Primary RX buffer allocation failed\n");
			kmem_cache_free(pinfo_cache, p_info);
			MSG_RTR_ARM_STAT_INC(rx2_q0_alloc_fail);
			return NULL;
		}
		p_info->primary_bufptr = bufptr;

		/* Same as skb_reserve(skb, msg_rtr_arm->rx_reserve_headroom) */
		data = bufptr + msg_rtr_arm->rx_vtc.rx_reserve_headroom;

		sg_init_table(p_info->sg, MSG_RTR_ARM_SGLIST_SIZE);
		sg_set_buf(&p_info->sg[0], data, size);

		p_info->sg_ents = dma_map_sg(RX_DEVICE, &p_info->sg[0],
			1, DMA_FROM_DEVICE);
		if (p_info->sg_ents != 1) {
			dev_err(RX_DEVICE, "vtc: dma map failed\n");
			msg_rtr_arm_frag_free((bufsiz <= PAGE_SIZE), bufptr);
			kmem_cache_free(pinfo_cache, p_info);
			MSG_RTR_ARM_STAT_INC(rx2_q0_alloc_fail);
			return NULL;
		}

		desc = dmaengine_prep_slave_sg(msg_rtr_arm->rx_vtc.rx_channel, p_info->sg,
			1, DMA_DEV_TO_MEM, 0);
		if (IS_ERR_OR_NULL(desc)) {
			dma_unmap_sg(RX_DEVICE, &p_info->sg[0], 1, DMA_FROM_DEVICE);
			msg_rtr_arm_frag_free((bufsiz <= PAGE_SIZE), bufptr);
			kmem_cache_free(pinfo_cache, p_info);
			err = PTR_ERR(desc);
			if (err != -ENOMEM) {
				dev_err(RX_DEVICE,
					"vtc: dma prep failed, error %d\n", err);
			}
			MSG_RTR_ARM_STAT_INC(rx2_q0_alloc_fail);
			return NULL;
		}

		desc->callback_param = p_info;
		desc->callback = msg_rtr_arm_rx_complete_vtc;
		p_info->cookie = desc->cookie;

		MSG_RTR_ARM_STAT_INC(rx2_q0_alloc);

	} else {
		struct page *page;

		/* Allocate a secondary receive queue entry */
		struct scatterlist sg[1];

		// DBG("vtc: msg_rtr_arm Attempting to alloc in pool 1");

		page = alloc_page(GFP_ATOMIC | GFP_DMA32 | __GFP_COLD);
		if (!page) {
			dev_warn(RX_DEVICE, "page alloc failed for pool %d\n", q_num);
			MSG_RTR_ARM_STAT_INC(rx2_q1_alloc_fail);
			return NULL;
		}

		sg_init_table(sg, 1);
		sg_set_page(&sg[0], page, PAGE_SIZE, 0);

		err = dma_map_sg(RX_DEVICE, sg, 1, DMA_FROM_DEVICE);
		if (err != 1) {
			dev_warn(RX_DEVICE, "vtc: map error for pool %d\n", q_num);
			__free_page(page);
			MSG_RTR_ARM_STAT_INC(rx2_q1_alloc_fail);
			return NULL;
		}

		desc = dmaengine_prep_slave_sg(msg_rtr_arm->rx_vtc.rx_channel, sg, 1,
			DMA_DEV_TO_MEM,
			q_num << DMA_QNUM_SHIFT);
		if (IS_ERR_OR_NULL(desc)) {
			dma_unmap_sg(RX_DEVICE, sg, 1, DMA_FROM_DEVICE);
			__free_page(page);

			err = PTR_ERR(desc);
			if (err != -ENOMEM) {
				dev_err(RX_DEVICE,
					"vtc: dma prep failed, error %d\n", err);
			}
			MSG_RTR_ARM_STAT_INC(rx2_q1_alloc_fail);
			return NULL;
		}

		desc->callback_param = page;
		desc->callback = msg_rtr_arm_rx_complete2nd;

		MSG_RTR_ARM_STAT_INC(rx2_q1_alloc);
	}

	wmb();
	return desc;
}

/*
 *  Another ugly layering violation. We can only pass one
 *  value describing a secondary buffer, but we need both
 *  the page and dma addresses. -rrp
 */
unsigned long	__arch_dma_pfn_offset = PFN_DOWN(KEYSTONE_HIGH_PHYS_START - KEYSTONE_LOW_PHYS_START);
#define	page_to_dma(dev,pg)	pfn_to_dma(dev, page_to_pfn(pg))

/* Release a free receive buffer */
static void msg_rtr_arm_rxpool_free(void *arg, unsigned q_num, unsigned bufsize,
	struct dma_async_tx_descriptor *desc)
{
	struct msg_rtr_arm_priv *msg_rtr_arm = arg;
        struct msg_rtr_arm_rx_struct* rx_struct = &msg_rtr_arm->rx_cp;

	CALLTRACE();

	if (q_num == 0) {
		struct msg_rtr_arm_packet *p_info = desc->callback_param;

		dma_unmap_single(RX_DEVICE, sg_dma_address(&p_info->sg[0]),
			p_info->primary_bufsiz, DMA_FROM_DEVICE);
		msg_rtr_arm_frag_free((p_info->primary_bufsiz <= PAGE_SIZE),
			p_info->primary_bufptr);
		kmem_cache_free(pinfo_cache, p_info);
	} else {
		struct page *page = desc->callback_param;
		dma_unmap_page(RX_DEVICE, page_to_dma(RX_DEVICE, page), // TODO double check with huge packets
			PAGE_SIZE, DMA_FROM_DEVICE);
		__free_page(page);
	}
}

/* Release a free receive buffer */
static void msg_rtr_arm_rxpool_free2(void *arg, unsigned q_num, unsigned bufsize,
	struct dma_async_tx_descriptor *desc)
{
	struct msg_rtr_arm_priv *msg_rtr_arm = arg;
        struct msg_rtr_arm_rx_struct* rx_struct = &msg_rtr_arm->rx_vtc;

	CALLTRACE();

	if (q_num == 0) {
		struct msg_rtr_arm_packet *p_info = desc->callback_param;

		dma_unmap_single(RX_DEVICE, sg_dma_address(&p_info->sg[0]),
			p_info->primary_bufsiz, DMA_FROM_DEVICE);
		msg_rtr_arm_frag_free((p_info->primary_bufsiz <= PAGE_SIZE),
			p_info->primary_bufptr);
		kmem_cache_free(pinfo_cache, p_info);
	} else {
		struct page *page = desc->callback_param;
		dma_unmap_page(RX_DEVICE, page_to_dma(RX_DEVICE, page), // TODO double check with huge packets
			PAGE_SIZE, DMA_FROM_DEVICE);
		__free_page(page);
	}
}

static int msg_rtr_arm_tx_unmap_skb(struct device *dev, struct sk_buff *skb, int offset, int len)
{
	int start = skb_headlen(skb);
	int i, copy = start - offset;
	struct sk_buff *frag_iter;
	int elt = 0;

	CALLTRACE();

	if (copy > 0) {
		if (copy > len)
			copy = len;
		dma_unmap_single(dev, virt_to_dma(dev, skb->data + offset),
				copy, DMA_TO_DEVICE);
		elt++;
		if ((len -= copy) == 0)
			return elt;
		offset += copy;
	}

	for (i = 0; i < skb_shinfo(skb)->nr_frags; i++) {
		void *virt_addr;
		int end;

		WARN_ON(start > offset + len);

		end = start + skb_frag_size(&skb_shinfo(skb)->frags[i]);
		if ((copy = end - offset) > 0) {
			skb_frag_t *frag = &skb_shinfo(skb)->frags[i];

			if (copy > len)
				copy = len;
			virt_addr = skb_frag_address(frag) + offset - start;
			dma_unmap_page(dev, virt_to_dma(dev, virt_addr),
					copy, DMA_TO_DEVICE);
			elt++;
			if (!(len -= copy))
				return elt;
			offset += copy;
		}
		start = end;
	}

	skb_walk_frags(skb, frag_iter) {
		int end;

		WARN_ON(start > offset + len);

		end = start + frag_iter->len;
		if ((copy = end - offset) > 0) {
			if (copy > len)
				copy = len;
			elt += msg_rtr_arm_tx_unmap_skb(dev, frag_iter,
						offset - start, copy);
			if ((len -= copy) == 0)
				return elt;
			offset += copy;
		}
		start = end;
	}
	BUG_ON(len);
	return elt;
}

static void msg_rtr_arm_tx_complete(void *data, struct msg_rtr_arm_tx_struct* tx_struct)
{
	struct sk_buff *skb = data;
	//struct msg_rtr_arm_priv *msg_rtr_arm = msg_rtr_arm_global_data.msg_rtr_arm;
	unsigned int ents;

	CALLTRACE();

	ents = msg_rtr_arm_tx_unmap_skb(TX_DEVICE, skb, 0, skb->len);
	atomic_add_return(ents, &tx_struct->tx_poll_count);
	dev_kfree_skb(skb);
}

static void msg_rtr_arm_tx_complete_cp(void *data)
{
	struct msg_rtr_arm_priv *msg_rtr_arm = msg_rtr_arm_global_data.msg_rtr_arm;

	CALLTRACE();

	msg_rtr_arm_tx_complete(data, &msg_rtr_arm->tx_cp);
}

static void msg_rtr_arm_tx_complete_vtc(void *data)
{
	struct msg_rtr_arm_priv *msg_rtr_arm = msg_rtr_arm_global_data.msg_rtr_arm;

	CALLTRACE();

	msg_rtr_arm_tx_complete(data, &msg_rtr_arm->tx_vtc);
}

static inline unsigned int msg_rtr_arm_find_dest_flow(unsigned int core)
{
	//unsigned int i;

	CALLTRACE();
        if(core2flowmap[core] > 0) {
            return core2flowmap[core];
        }
#if 0
        // DSPs and CCS_RT
	for (i = 0; i < MAX_DSPS+2; i++)
	{
		if (msg_rtr_arm_flow_map[i].core == core)
		{
			return msg_rtr_arm_flow_map[i].flow;
		}
	}
#endif
	return ERR_FLOW;
}

static inline int msg_rtr_arm_send_dsp(struct sk_buff *skb, u32 flowtag, struct msg_rtr_arm_tx_struct* tx_struct, u32 coreId)
{
	struct dma_async_tx_descriptor *desc;
	struct msg_rtr_arm_packet p_info;
	int real_sg_ents;
	int ret = 0;
	dma_cookie_t cookie;
#ifdef MSG_RTR_ARM_INLINE_TX_RECYCLE
	u32 poll_count;
#endif
        unsigned long flags = 0ul;

	CALLTRACE();

	if (unlikely(flowtag > DMA_FLOWTAG_MASK))
	{
		MSG_RTR_ARM_STAT_INC(tx_dropped_bad_dest);
		dev_warn(TX_DEVICE, "Invalid destination flow: %u\n", flowtag);
		ret = -ENXIO;
		goto drop;
	}

	sg_init_table(p_info.sg, MSG_RTR_ARM_SGLIST_SIZE);

        if(likely(coreId == 0xA)) {
            u32 epib[4] = {0, 0x0000000Au, 0, 0};
            sg_set_buf(&p_info.sg[0], epib, sizeof(epib));
	    real_sg_ents = skb_to_sgvec(skb, &p_info.sg[1], 0, skb->len);
            real_sg_ents++;
            flags = DMA_HAS_EPIB | DMA_HAS_FLOWTAG | ((flowtag & DMA_FLOWTAG_MASK) << DMA_FLOWTAG_SHIFT);
            DBG("msg_rtr_arm_send_dsp core=A, flags=%x", (unsigned int)flags);
        } else if(coreId == 0x9) {
            u32 epib[4] = {0, 0x00000009u, 0, 0};
            sg_set_buf(&p_info.sg[0], epib, sizeof(epib));
	    real_sg_ents = skb_to_sgvec(skb, &p_info.sg[1], 0, skb->len);
            real_sg_ents++;
            flags = DMA_HAS_EPIB | DMA_HAS_FLOWTAG | ((flowtag & DMA_FLOWTAG_MASK) << DMA_FLOWTAG_SHIFT);
            DBG("msg_rtr_arm_send_dsp core=9. flags=%x", (unsigned int)flags);
        } else {
	    real_sg_ents = skb_to_sgvec(skb, &p_info.sg[0], 0, skb->len);
            flags =  DMA_HAS_FLOWTAG | ((flowtag & DMA_FLOWTAG_MASK) << DMA_FLOWTAG_SHIFT);
            DBG("msg_rtr_arm_send_dsp core=OTHER, flags=%x", (unsigned int)flags);
        }

	/* Map all the packet fragments	into the scatterlist */
	p_info.sg_ents = dma_map_sg(TX_DEVICE, &p_info.sg[0], real_sg_ents, DMA_TO_DEVICE);
	if (p_info.sg_ents != (real_sg_ents)) {
		MSG_RTR_ARM_STAT_INC(tx_dropped_map_fail);
		dev_warn(TX_DEVICE, "failed to map transmit packet\n");
		ret = -ENXIO;
		goto drop;
	}
#ifdef MSG_RTR_ARM_INLINE_TX_RECYCLE
	poll_count = atomic_sub_return(real_sg_ents, &tx_struct->tx_poll_count);
	if (poll_count < (tx_struct->tx_queue_depth / 2) ) {
		dmaengine_pause(tx_struct->tx_channel);
		DBG("Forcing Poll");
		dma_poll(tx_struct->tx_channel, TX_POLL_BUDGET);
		dmaengine_resume(tx_struct->tx_channel);
	}
	poll_count = atomic_read(&tx_struct->tx_poll_count);
#endif

	desc = dmaengine_prep_slave_sg(tx_struct->tx_channel, p_info.sg, p_info.sg_ents,
                                       DMA_MEM_TO_DEV, flags);

	if (IS_ERR_OR_NULL(desc)) {
		MSG_RTR_ARM_STAT_INC(tx_dropped_prep_fail);
		dma_unmap_sg(TX_DEVICE, &p_info.sg[0], real_sg_ents,
			DMA_TO_DEVICE);
		dev_dbg(TX_DEVICE, "failed to prep slave dma %ld\n", PTR_ERR(desc));
#ifdef MSG_RTR_ARM_INLINE_TX_RECYCLE
		if (-ENOMEM == PTR_ERR(desc)) {
			dmaengine_pause(tx_struct->tx_channel);
			DBG("Forcing Poll");
			dma_poll(tx_struct->tx_channel, TX_POLL_BUDGET);
			dmaengine_resume(tx_struct->tx_channel);
		}
#endif
		ret = -ENOBUFS;
		goto drop;
	}

	desc->callback_param = skb;
	desc->callback = tx_struct->callback;
	wmb();

	MSG_RTR_ARM_STAT_INC(tx_dsp_pkt);

	cookie = dmaengine_submit(desc);
	if (dma_submit_error(cookie)) {
		dev_warn(TX_DEVICE, "failed to submit packet for dma: %d\n", cookie);
		MSG_RTR_ARM_STAT_INC(tx_dropped_submit_fail);
		goto drop;
	}
	DBG("Submitted packet");
	return MSG_RTR_ARM_TX_OK;

drop:
	DBG("Dropped packet");
	MSG_RTR_ARM_STAT_INC(tx_dropped_total);
	// dev_kfree_skb(skb); // Based on API doc we just return failure without freeing
	return ret;
}

static inline int msg_rtr_arm_send_internal(struct sk_buff *skb, bool lb, struct msg_rtr_arm_tx_struct* tx_struct)
{
	struct msg_rtr_arm_priv *msg_rtr_arm = msg_rtr_arm_global_data.msg_rtr_arm;
	TAaSysComSicad tmpSicad = 0;
	struct SMsgData *sicad;
	u32 flowtag;
        u32 coreId = 0;

	CALLTRACE();

	if (skb->len < sizeof(SInternalMsgHeader))
	{
		MSG_RTR_ARM_STAT_INC(tx_dropped_internal);
		dev_dbg(TX_DEVICE, "packet size is too small to be a real Syscom Message: %u\n", skb->len);
		return -1;
	}

	//swap the source and destination SICAD
	sicad = (struct SMsgData*)skb->data;

	tmpSicad = sicad->header.target;
	if (unlikely(lb)) {
		sicad->header.target = sicad->header.sender;
		sicad->header.sender = tmpSicad;
		flowtag = msg_rtr_arm->dma_flow_loopback;
	} else {
                coreId = (tmpSicad >> 16) & 0x000F;
		flowtag = msg_rtr_arm_find_dest_flow(coreId);
		DBG("Determined flow from target Sicad: %x to be %u\n", tmpSicad, flowtag);
	}

	if (flowtag > DMA_FLOWTAG_MASK)
	{
		//not a valid DSP, send to netlink
		return msg_rtr_arm_netlink_classify_and_send(skb, tmpSicad >> 16);
	}
	else
	{
		return msg_rtr_arm_send_dsp(skb, flowtag, tx_struct, coreId);
	}

}

int MsgRtrARMSend_loopback(struct sk_buff *skb)
{

	CALLTRACE();

	MSG_RTR_ARM_STAT_INC(rx_vtc_pkt);
	return msg_rtr_arm_send_internal(skb, true, &msg_rtr_arm_global_data.msg_rtr_arm->tx_vtc);
}
EXPORT_SYMBOL(MsgRtrARMSend_loopback);

int MsgRtrARMSend(struct sk_buff *skb)
{

	CALLTRACE();

	MSG_RTR_ARM_STAT_INC(rx_vtc_pkt);
	if (msg_rtr_arm_global_data.msg_rtr_arm->tx_trace) {
		MSG_RTR_ARM_STAT_INC(tx_trace_pkt);
		msg_rtr_arm_send_netlink_msg(msg_rtr_arm_global_data.msg_rtr_arm,skb_clone(skb, GFP_ATOMIC),msg_rtr_arm_global_data.msg_rtr_arm->nl_linux_pid,true);
	}
	return msg_rtr_arm_send_internal(skb, false, &msg_rtr_arm_global_data.msg_rtr_arm->tx_vtc);
}
EXPORT_SYMBOL(MsgRtrARMSend);

static int msg_rtr_arm_open_rx_cp(struct msg_rtr_arm_priv *msg_rtr_arm);
static int msg_rtr_arm_open_rx_vtc(struct msg_rtr_arm_priv *msg_rtr_arm);

/* Open the receive channel */
static int msg_rtr_arm_open_rx(struct msg_rtr_arm_priv *msg_rtr_arm) {
    int ret = 0;
    ret = msg_rtr_arm_open_rx_cp(msg_rtr_arm);
    if (ret < 0) {
        ERR("msg_rtr_arm open rx_cp failed");
        return ret;
    }
    
    ret = msg_rtr_arm_open_rx_vtc(msg_rtr_arm);
    if (ret < 0) {
        ERR("msg_rtr_arm open rx_vtc failed");
        msg_rtr_arm_close_channel(msg_rtr_arm->rx_cp.rx_channel);
        msg_rtr_arm->rx_cp.rx_channel = NULL;
        return ret;
    }
    return 0;
}

static int msg_rtr_arm_open_rx_cp(struct msg_rtr_arm_priv *msg_rtr_arm)
{
	struct dma_keystone_info config;
	dma_cap_mask_t mask;
	int err = -ENODEV;
	const char *name;
	int i;
	/* u32 reg_ptr; */
        struct msg_rtr_arm_rx_struct* rx_struct = &msg_rtr_arm->rx_cp;

	CALLTRACE();

	dma_cap_zero(mask);
	dma_cap_set(DMA_SLAVE, mask);

	name = msg_rtr_arm->rx_cp.rx_chan_name;
	msg_rtr_arm->rx_cp.rx_channel = dma_request_channel_by_name(mask, name);
	if (IS_ERR_OR_NULL(msg_rtr_arm->rx_cp.rx_channel))
        {
		return err;
        }

	memset(&config, 0, sizeof(config));
	config.direction                = DMA_DEV_TO_MEM;
	config.scatterlist_size         = MSG_RTR_ARM_SGLIST_SIZE;
	config.rxpool_allocator         = msg_rtr_arm_rxpool_alloc;
	config.rxpool_destructor        = msg_rtr_arm_rxpool_free;
	config.rxpool_param             = msg_rtr_arm;
	config.rxpool_thresh_enable     = DMA_THRESH_NONE;

	for (i = 0; i < MSG_RTR_ARM_NUM_RX_QUEUES &&
		msg_rtr_arm->rx_cp.rx_queue_depths[i] &&
		msg_rtr_arm->rx_cp.rx_buffer_sizes[i]; ++i) {
		config.rxpools[i].pool_depth  = msg_rtr_arm->rx_cp.rx_queue_depths[i];
		config.rxpools[i].buffer_size = msg_rtr_arm->rx_cp.rx_buffer_sizes[i];
		dev_dbg(RX_DEVICE, "rx_pool[%d] depth %d, size %d\n", i,
			config.rxpools[i].pool_depth,
			config.rxpools[i].buffer_size);
	}
	config.rxpool_count = i;
	err = dma_keystone_config(msg_rtr_arm->rx_cp.rx_channel, &config);

	if (err) {
		dev_err(RX_DEVICE, "%d error configuring RX channel\n",
			err);
		goto fail;
	}
	dev_dbg(RX_DEVICE, "opened RX channel: %p\n", msg_rtr_arm->rx_cp.rx_channel);

#if 0
	if (msg_rtr_arm->rx_cp.rx_free_q0 >= LEAST_STARVATION_QUEUE &&
		msg_rtr_arm->rx_cp.rx_free_q0 <= GREATEST_STARVATION_QUEUE) // TODO error case
	{
		reg_ptr = QUEUE_CONFIG_BASE_REGISTER + 0x20 + ((msg_rtr_arm->rx_cp.rx_free_q0 - LEAST_STARVATION_QUEUE) >> 2) * 4;
		dev_err(RX_DEVICE, "preparing to rx_cp request_mem_region for %x\n", reg_ptr);

		if (!request_mem_region(reg_ptr, 1, "Starvation Queues"))
		{
			dev_err(RX_DEVICE, "Unable to reserve memory resource for rx_cp mra\n");
			err = -1;
			goto fail;
		}

		msg_rtr_arm->rx_cp.rx_free_reg_0 = ioremap_nocache(reg_ptr, 1);
		if (!msg_rtr_arm->rx_cp.rx_free_reg_0)
		{
			dev_err(RX_DEVICE, "Unable to map register for starvartion queue\n");
			release_mem_region(reg_ptr,1);
			err = -1;
			goto fail;
		}
	}
#endif

	dma_set_notify(msg_rtr_arm->rx_cp.rx_channel, msg_rtr_arm_rx_notify, &msg_rtr_arm->rx_cp);

	dma_rxfree_refill(msg_rtr_arm->rx_cp.rx_channel);

	return 0;

fail:
	if (msg_rtr_arm->rx_cp.rx_channel) {
		dma_release_channel(msg_rtr_arm->rx_cp.rx_channel);
		msg_rtr_arm->rx_cp.rx_channel = NULL;
	}
	return err;
}

static int msg_rtr_arm_open_rx_vtc(struct msg_rtr_arm_priv *msg_rtr_arm)
{
	struct dma_keystone_info config;
	dma_cap_mask_t mask;
	int err = -ENODEV;
	const char *name;
	int i;
	/* u32 reg_ptr; */
        struct msg_rtr_arm_rx_struct* rx_struct = &msg_rtr_arm->rx_vtc;

	CALLTRACE();

	dma_cap_zero(mask);
	dma_cap_set(DMA_SLAVE, mask);

	name = msg_rtr_arm->rx_vtc.rx_chan_name;
	msg_rtr_arm->rx_vtc.rx_channel = dma_request_channel_by_name(mask, name);
	if (IS_ERR_OR_NULL(msg_rtr_arm->rx_vtc.rx_channel))
		return err;

	memset(&config, 0, sizeof(config));
	config.direction                = DMA_DEV_TO_MEM;
	config.scatterlist_size         = MSG_RTR_ARM_SGLIST_SIZE;
	config.rxpool_allocator         = msg_rtr_arm_rxpool_alloc2;
	config.rxpool_destructor        = msg_rtr_arm_rxpool_free2;
	config.rxpool_param             = msg_rtr_arm;
	config.rxpool_thresh_enable     = DMA_THRESH_NONE;

	for (i = 0; i < MSG_RTR_ARM_NUM_RX_QUEUES &&
		msg_rtr_arm->rx_vtc.rx_queue_depths[i] &&
		msg_rtr_arm->rx_vtc.rx_buffer_sizes[i]; ++i) {
		config.rxpools[i].pool_depth  = msg_rtr_arm->rx_vtc.rx_queue_depths[i];
		config.rxpools[i].buffer_size = msg_rtr_arm->rx_vtc.rx_buffer_sizes[i];
		dev_dbg(RX_DEVICE, "vtc: rx_pool[%d] depth %d, size %d\n", i,
			config.rxpools[i].pool_depth,
			config.rxpools[i].buffer_size);
	}
	config.rxpool_count = i;
	err = dma_keystone_config(msg_rtr_arm->rx_vtc.rx_channel, &config);

	if (err) {
		dev_err(RX_DEVICE, "vtc: %d error configuring RX channel\n",
			err);
		goto fail;
	}
	dev_dbg(RX_DEVICE, "opened RX channel: %p\n", msg_rtr_arm->rx_vtc.rx_channel);

#if 0
	if (msg_rtr_arm->rx_vtc.rx_free_q0 >= LEAST_STARVATION_QUEUE &&
		msg_rtr_arm->rx_vtc.rx_free_q0 <= GREATEST_STARVATION_QUEUE) // TODO error case
	{
		reg_ptr = QUEUE_CONFIG_BASE_REGISTER + 0x20 + ((msg_rtr_arm->rx_vtc.rx_free_q0 - LEAST_STARVATION_QUEUE) >> 2) * 4;
                dev_err(RX_DEVICE, "preparing to rx_vtc request_mem_region for %x\n", reg_ptr);

		if (!request_mem_region(reg_ptr, 1, "Starvation Queues"))
		{
			dev_err(RX_DEVICE, "Unable to reserve memory resource for rx_vtc mra\n");
			err = -1;
			goto fail;
		}

		msg_rtr_arm->rx_vtc.rx_free_reg_0 = ioremap_nocache(reg_ptr, 1);
		if (!msg_rtr_arm->rx_vtc.rx_free_reg_0)
		{
			dev_err(RX_DEVICE, "vtc: Unable to map register for starvartion queue\n");
			release_mem_region(reg_ptr,1);
			err = -1;
			goto fail;
		}
	}
#endif

	dma_set_notify(msg_rtr_arm->rx_vtc.rx_channel, msg_rtr_arm_rx_notify, &msg_rtr_arm->rx_vtc);

	dma_rxfree_refill(msg_rtr_arm->rx_vtc.rx_channel);

	return 0;

fail:
	if (msg_rtr_arm->rx_vtc.rx_channel) {
		dma_release_channel(msg_rtr_arm->rx_vtc.rx_channel);
		msg_rtr_arm->rx_vtc.rx_channel = NULL;
	}
	return err;
}

static void msg_rtr_arm_rcv_netlink_msg(struct sk_buff * skb)
{
	struct nlmsghdr *nlh;
	struct sk_buff * skb_clone_buf;
	struct msg_rtr_arm_priv * msg_rtr_arm = msg_rtr_arm_global_data.msg_rtr_arm;
	ddal_k2mr_msg_rtr_arm_header_t * k2mr_hdr;
	nlh = (struct nlmsghdr*)skb->data;

	CALLTRACE();

	//We need to add support for hwapi restarting, according to Andy
	/*the first message will populate the pid*/

	skb_clone_buf = skb_clone(skb, GFP_ATOMIC);
	skb_pull(skb_clone_buf, NLMSG_HDRLEN);

	k2mr_hdr = (ddal_k2mr_msg_rtr_arm_header_t *)skb_clone_buf->data;

	DBG("len: %u, clonelen: %u, pid: %u, type: %u, flags: %u, skbptr: %p, skbclone: %p", skb->len, skb_clone_buf->len, nlh->nlmsg_pid, k2mr_hdr->type, k2mr_hdr->flags, skb, skb_clone_buf);

	if (unlikely(DDAL_IGNORE_CONTROL_TYPE != k2mr_hdr->type))
	{
		MSG_RTR_ARM_STAT_INC(rx_nl_ctrl_pkt);
		switch (k2mr_hdr->type) {

		case DDAL_REGISTER :
			DBG("received DDAL_REGISTER len:%u pid:%u", skb->len, msg_rtr_arm->nl_linux_pid);
			msg_rtr_arm->nl_linux_pid = nlh->nlmsg_pid;
			break;
		case DDAL_ENABLE_LOOPBACK :
			DBG("received DDAL_ENABLE_LOOPBACK");
			msg_rtr_arm->loopback = true;
			break;
		case DDAL_DISABLE_LOOPBACK :
			msg_rtr_arm->loopback = false;
			DBG("received DDAL_DISABLE_LOOPBACK");
			break;
		case DDAL_ENABLE_TRACE_RX:
			DBG("received DDAL_ENABLE_TRACE_RX");
			msg_rtr_arm->rx_trace = true;
			break;
		case DDAL_DISABLE_TRACE_RX:
			DBG("received DDAL_DISABLE_TRACE_RX");
			msg_rtr_arm->rx_trace = false;
			break;
		case DDAL_ENABLE_TRACE_TX :
			DBG("received DDAL_ENABLE_TRACE_TX");
			msg_rtr_arm->tx_trace = true;
			break;
		case DDAL_DISABLE_TRACE_TX :
			DBG("received DDAL_DISABLE_TRACE_TX");
			msg_rtr_arm->tx_trace = false;
			break;
		case DDAL_REGISTER_NID:
			DBG("received DDAL_REGISTER_NID nid:%x pid:%u", k2mr_hdr->flags, nlh->nlmsg_pid);
			msg_rtr_arm_register_netlink_pid(k2mr_hdr->flags, nlh->nlmsg_pid);
			break;
		default :
			DBG("received UNKNOWN Control Message: %u", k2mr_hdr->type);
			break;
		}
		dev_kfree_skb_any(skb_clone_buf);
		return;
	}

	skb_pull(skb_clone_buf, sizeof(ddal_k2mr_msg_rtr_arm_header_t)); //This should be the size of shared header
	DBG("rcv_netlink_message skb size after pull: %u", skb_clone_buf->len);
	MSG_RTR_ARM_STAT_INC(rx_nl_pkt);
	if (msg_rtr_arm_send_internal(skb_clone_buf, msg_rtr_arm->loopback,
		&msg_rtr_arm_global_data.msg_rtr_arm->tx_cp) < 0)
	{
		dev_kfree_skb_any(skb_clone_buf);  //msg_rtr_arm_send_internal does not free the skb on a send failure, we do it here
	}
}

static int msg_rtr_arm_register_netlink(struct msg_rtr_arm_priv *msg_rtr_arm)
{
	struct netlink_kernel_cfg cfg;

	CALLTRACE();

	cfg.groups = 0; //No idea what to do here
	cfg.input = msg_rtr_arm_rcv_netlink_msg;
	cfg.cb_mutex = NULL; //Do we need a mutex?
	cfg.bind = NULL; //bind function?  wtf is this?

	//TODO, get rid of magic number 24
	msg_rtr_arm->nl_sock = netlink_kernel_create(&init_net,/* NETLINK_FZM*/  24, &cfg);
	if (msg_rtr_arm->nl_sock == NULL) {
		return -ENOMEM;
	}
	return 0;
}

#define MSG_RTR_ARM_STRING(str) \
	do { \
	len = snprintf(buf, count+1-total_len, "%s", str); \
	if (len > count-total_len) len = count-total_len; \
	total_len += len; \
	buf += len; \
	} while (0)

#define MSG_RTR_ARM_PROC_STAT(title, field) \
	do { \
	len = snprintf(buf, count+1-total_len, "%-40s\t%u\n", (title), msg_rtr_arm_global_data.msg_rtr_arm->stats.field); \
	if (len > count-total_len) len = count-total_len; \
	total_len += len; \
	buf += len; \
	} while (0)

#define MSG_RTR_ARM_PROC_STAT_RAW(field) \
	do { \
	len = snprintf(buf, count+1-total_len, "%u\n", msg_rtr_arm_global_data.msg_rtr_arm->stats.field); \
	if (len > count-total_len) len = count-total_len; \
	total_len += len; \
	buf += len; \
	} while (0)


#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
static ssize_t msg_rtr_arm_read_proc(struct file*  file_,
                                    char __user*  page,
                                    size_t        count,
                                    loff_t *      offset_)
#else /* LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0) */
static int msg_rtr_arm_read_proc(char *page, char **start, off_t off,
	int count, int *eof, void *data)
#endif
{
	int total_len = 0;
	int len;
	char *buf, *buf0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
	if (*offset_ != 0)
		return 0;
#endif
	buf0 = buf = kmalloc(count+1, GFP_KERNEL); /* including '\0' */
	if (!buf)
		return -ENOMEM;

	MSG_RTR_ARM_STRING("## MsgRtrARM Stats ##\n");



	MSG_RTR_ARM_PROC_STAT("Total Packets Received From DSP:", rx_dsp_pkt);
	MSG_RTR_ARM_PROC_STAT("Total Packets Received From Netlink:", rx_nl_pkt);
	MSG_RTR_ARM_PROC_STAT("Total Control Received From Netlink:", rx_nl_ctrl_pkt);
	MSG_RTR_ARM_PROC_STAT("Total Packets Received From VTC:", rx_vtc_pkt);
	MSG_RTR_ARM_PROC_STAT("Total Packets Sent to DSPs:", tx_dsp_pkt);
	MSG_RTR_ARM_PROC_STAT("Total Packets Sent to Netlink:", tx_nl_pkt);
	MSG_RTR_ARM_PROC_STAT("Total Packets Sent to VTC:", tx_vtc_pkt);
	MSG_RTR_ARM_PROC_STAT("Total Receive Interrupts:", rx_int);
	MSG_RTR_ARM_PROC_STAT("Buffers Allocated in Q0:", rx_q0_alloc);
	MSG_RTR_ARM_PROC_STAT("Buffers Allocated in Q1:", rx_q1_alloc);
	MSG_RTR_ARM_PROC_STAT("Transmit Packets Mirrored:", tx_trace_pkt);
	MSG_RTR_ARM_PROC_STAT("Receive Packets Mirrored:", rx_trace_pkt);
	MSG_RTR_ARM_PROC_STAT("Buffer Allocation Failures for Q0:", rx_q0_alloc_fail);
	MSG_RTR_ARM_PROC_STAT("Buffer Allocation Failures for Q1:", rx_q1_alloc_fail);
	MSG_RTR_ARM_PROC_STAT("Receive DMA Failures detected:", rx_dma_fail);
	MSG_RTR_ARM_PROC_STAT("Transmit DMA Failures detected:", tx_dma_fail);
	MSG_RTR_ARM_PROC_STAT("Transmit Netlink Failures detected:", tx_nl_fail);
	MSG_RTR_ARM_PROC_STAT("Transmit DMA Dropped (bad destination):", tx_dropped_bad_dest);
	MSG_RTR_ARM_PROC_STAT("Transmit DMA Dropped (map failed):",tx_dropped_map_fail);
	MSG_RTR_ARM_PROC_STAT("Transmit DMA Dropped (prep slave failed):",tx_dropped_prep_fail);
	MSG_RTR_ARM_PROC_STAT("Transmit DMA Dropped (submit failed):",tx_dropped_submit_fail);
	MSG_RTR_ARM_PROC_STAT("Transmit DMA Dropped (internal):",tx_dropped_internal);
	MSG_RTR_ARM_PROC_STAT("Transmit DMA Dropped (total):",tx_dropped_total);
	MSG_RTR_ARM_PROC_STAT("Receive Q0 Starvation Events:", rx_q0_starvation);
	MSG_RTR_ARM_PROC_STAT("Receive Q1 Starvation Events:", rx_q1_starvation);

	MSG_RTR_ARM_PROC_STAT("Buffers Allocated in Q0[vtc]:", rx2_q0_alloc);
	MSG_RTR_ARM_PROC_STAT("Buffers Allocated in Q1[vtc]:", rx2_q1_alloc);
	MSG_RTR_ARM_PROC_STAT("Buffer Allocation Failures for Q0[vtc]:", rx2_q0_alloc_fail);
	MSG_RTR_ARM_PROC_STAT("Buffer Allocation Failures for Q1[vtc]:", rx2_q1_alloc_fail);
	MSG_RTR_ARM_PROC_STAT("Receive Q0 Starvation Events[vtc]:", rx2_q0_starvation);
	MSG_RTR_ARM_PROC_STAT("Receive Q1 Starvation Events[vtc]:", rx2_q1_starvation);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
	*offset_ = total_len;

	if (copy_to_user(page, buf0, total_len)) {
		kfree(buf0);
		return -EACCES;
	}
#else /* LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0) */
	memcpy(page, buf0, total_len);
#endif

	kfree(buf0);

	return total_len;
}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
static ssize_t msg_rtr_arm_read_proc_raw(struct file*  file_,
                                         char __user*  page,
                                         size_t        count,
                                         loff_t *      offset_)
#else /* LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0) */
static int msg_rtr_arm_read_proc_raw(char *page, char **start, off_t off,
	int count, int *eof, void *data)
#endif
{
	int total_len = 0;
	int len;
	char *buf, *buf0;

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
	if (*offset_ != 0)
		return 0;
#endif

	buf0 = buf = kmalloc(count+1, GFP_KERNEL); /* including '\0' */
	if (!buf)
		return -ENOMEM;

	MSG_RTR_ARM_PROC_STAT_RAW(rx_dsp_pkt);
	MSG_RTR_ARM_PROC_STAT_RAW(rx_nl_pkt);
	MSG_RTR_ARM_PROC_STAT_RAW(rx_nl_ctrl_pkt);
	MSG_RTR_ARM_PROC_STAT_RAW(rx_vtc_pkt);
	MSG_RTR_ARM_PROC_STAT_RAW(tx_dsp_pkt);
	MSG_RTR_ARM_PROC_STAT_RAW(tx_nl_pkt);
	MSG_RTR_ARM_PROC_STAT_RAW(tx_vtc_pkt);
	MSG_RTR_ARM_PROC_STAT_RAW(rx_int);
	MSG_RTR_ARM_PROC_STAT_RAW(rx_q0_alloc);
	MSG_RTR_ARM_PROC_STAT_RAW(rx_q1_alloc);
	MSG_RTR_ARM_PROC_STAT_RAW(tx_trace_pkt);
	MSG_RTR_ARM_PROC_STAT_RAW(rx_trace_pkt);
	MSG_RTR_ARM_PROC_STAT_RAW(rx_q0_alloc_fail);
	MSG_RTR_ARM_PROC_STAT_RAW(rx_q1_alloc_fail);
	MSG_RTR_ARM_PROC_STAT_RAW(rx_dma_fail);
	MSG_RTR_ARM_PROC_STAT_RAW(tx_dma_fail);
	MSG_RTR_ARM_PROC_STAT_RAW(tx_nl_fail);
	MSG_RTR_ARM_PROC_STAT_RAW(tx_dropped_bad_dest);
	MSG_RTR_ARM_PROC_STAT_RAW(tx_dropped_map_fail);
	MSG_RTR_ARM_PROC_STAT_RAW(tx_dropped_prep_fail);
	MSG_RTR_ARM_PROC_STAT_RAW(tx_dropped_submit_fail);
	MSG_RTR_ARM_PROC_STAT_RAW(tx_dropped_internal);
	MSG_RTR_ARM_PROC_STAT_RAW(tx_dropped_total);
	MSG_RTR_ARM_PROC_STAT_RAW(rx_q0_starvation);
	MSG_RTR_ARM_PROC_STAT_RAW(rx_q1_starvation);

	MSG_RTR_ARM_PROC_STAT_RAW(rx2_q0_alloc);
	MSG_RTR_ARM_PROC_STAT_RAW(rx2_q1_alloc);
	MSG_RTR_ARM_PROC_STAT_RAW(rx2_q0_alloc_fail);
	MSG_RTR_ARM_PROC_STAT_RAW(rx2_q1_alloc_fail);
	MSG_RTR_ARM_PROC_STAT_RAW(rx2_q0_starvation);
	MSG_RTR_ARM_PROC_STAT_RAW(rx2_q1_starvation);

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
	*offset_ = total_len;

	if (copy_to_user(page, buf0, total_len)) {
		kfree(buf0);
		return -EACCES;
	}
#else /* LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0) */
	memcpy(page, buf0, total_len);
#endif

	kfree(buf0);

	return total_len;
}


#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
static struct file_operations fops_rtr_arm = {
	.read  = msg_rtr_arm_read_proc,
};
static struct file_operations fops_rtr_arm_raw = {
	.read  = msg_rtr_arm_read_proc_raw,
};
#endif

static int msg_rtr_arm_start(void)
{
	struct msg_rtr_arm_priv * msg_rtr_arm;
	int ret = 0;

	CALLTRACE();

	/* Create a cache for these commonly-used structures */
	pinfo_cache = kmem_cache_create("msg_rtr_arm_pinfo_cache",
		sizeof(struct msg_rtr_arm_packet), sizeof(void *),
		SLAB_HWCACHE_ALIGN, NULL);
	if (!pinfo_cache)
		return -ENOMEM;

	DBG("Start");

	msg_rtr_arm_wq = create_workqueue("msg_rtr_arm_queue");
	if (!msg_rtr_arm_wq) {
		ERR("msg_rtr_arm create_workqueue error");
		kmem_cache_destroy(pinfo_cache);
		pinfo_cache = NULL;
		return -ENOMEM;
	}

	msg_rtr_arm = kzalloc(sizeof(*msg_rtr_arm), GFP_KERNEL);
	if (!msg_rtr_arm) {
		destroy_workqueue(msg_rtr_arm_wq);
		kmem_cache_destroy(pinfo_cache);
		pinfo_cache = NULL;
		return -ENOMEM;
	}
	msg_rtr_arm_global_data.msg_rtr_arm  = msg_rtr_arm;

	tasklet_init(&msg_rtr_arm->rx_cp.rx_poll_tasklet, msg_rtr_arm_rx_poll, 0);
        tasklet_init(&msg_rtr_arm->rx_vtc.rx_poll_tasklet, msg_rtr_arm_rx_poll2, 0);
#ifndef MSG_RTR_ARM_INLINE_TX_RECYCLE
	tasklet_init(&msg_rtr_arm->tx_cp.tx_poll_tasklet, msg_rtr_arm_tx_poll, (unsigned long)&msg_rtr_arm_global_data.msg_rtr_arm->tx_cp);
	tasklet_init(&msg_rtr_arm->tx_vtc.tx_poll_tasklet, msg_rtr_arm_tx_poll, (unsigned long)&msg_rtr_arm_global_data.msg_rtr_arm->tx_vtc);
#endif

	//TODO: possibly put this into a config file
	msg_rtr_arm->tx_cp.tx_chan_name = "msg-rtr-arm-tx-cp";
	msg_rtr_arm->tx_vtc.tx_chan_name = "msg-rtr-arm-tx-vtc";
	msg_rtr_arm->tx_cp.tx_queue_depth = 256;
	msg_rtr_arm->tx_vtc.tx_queue_depth = 1792;


	msg_rtr_arm->rx_cp.rx_chan_name = "msg-rtr-arm-rx-k2mr";
	msg_rtr_arm->rx_cp.rx_queue_depths[0] = SMALL_PACKETS_QUEUE_DEPTH;
	msg_rtr_arm->rx_cp.rx_queue_depths[1] = LARGE_PACKETS_QUEUE_DEPTH;
	msg_rtr_arm->rx_cp.rx_buffer_sizes[0] = 2048;
	msg_rtr_arm->rx_cp.rx_buffer_sizes[1] = 4096;
	msg_rtr_arm->rx_cp.rx_reserve_headroom = 128;

        // Linux receiving flow is 11
	//msg_rtr_arm->dma_flow_loopback = 11;

	msg_rtr_arm->rx_cp.rx_free_q0 = 738;
	msg_rtr_arm->rx_cp.rx_free_q1 = 739;

	msg_rtr_arm->rx_vtc.rx_chan_name = "msg-rtr-arm-rx";
	msg_rtr_arm->rx_vtc.rx_queue_depths[0] = SMALL_PACKETS_QUEUE_DEPTH;
	msg_rtr_arm->rx_vtc.rx_queue_depths[1] = LARGE_PACKETS_QUEUE_DEPTH;
	msg_rtr_arm->rx_vtc.rx_buffer_sizes[0] = 2048;
	msg_rtr_arm->rx_vtc.rx_buffer_sizes[1] = 4096;
	msg_rtr_arm->rx_vtc.rx_reserve_headroom = 128;

        // vtc receiving flow is 12
	msg_rtr_arm->dma_flow_loopback = 12;
	msg_rtr_arm->rx_vtc.rx_free_q0 = 736;
	msg_rtr_arm->rx_vtc.rx_free_q1 = 737;

        if(tst_loopback) {
            msg_rtr_arm->loopback = true;
            msg_rtr_arm->dma_flow_loopback = tst_loopback_flow;
        }

	ret = msg_rtr_arm_open_tx(msg_rtr_arm);
	if (ret < 0) {
		ERR("msg_rtr_arm Failed to open tx channel");
		msg_rtr_arm_close(msg_rtr_arm);
		return ret;
	}

	ret = msg_rtr_arm_open_rx(msg_rtr_arm);
	if (ret < 0) {
		ERR("msg_rtr_arm Failed to open rx channel");
		msg_rtr_arm_close(msg_rtr_arm);
		return ret;
	}

	ret = msg_rtr_arm_register_netlink(msg_rtr_arm);
	if (ret < 0) {
		ERR("msg_rtr_arm Failed to open netlink socket");
		msg_rtr_arm_close(msg_rtr_arm);
		return ret;
	}

#if LINUX_VERSION_CODE >= KERNEL_VERSION(3,10,0)
	msg_rtr_arm->proc_file = proc_create_data("msg_rtr_arm", 0444, NULL, &fops_rtr_arm, NULL);
	msg_rtr_arm->proc_file_raw = proc_create_data("msg_rtr_arm_raw", 0444, NULL, &fops_rtr_arm_raw, NULL);
#else /* LINUX_VERSION_CODE < KERNEL_VERSION(3,10,0) */
	msg_rtr_arm->proc_file = create_proc_read_entry("msg_rtr_arm", 0444, NULL, msg_rtr_arm_read_proc, NULL);
	msg_rtr_arm->proc_file_raw = create_proc_read_entry("msg_rtr_arm_raw", 0444, NULL, msg_rtr_arm_read_proc_raw, NULL);
#endif
	return 0;
}

static int __exit msg_rtr_arm_remove(void)
{

	CALLTRACE();

	return msg_rtr_arm_close(msg_rtr_arm_global_data.msg_rtr_arm);
}

static int __init msg_rtr_arm_init(void)
{
	int ret;

	CALLTRACE();

	ret = msg_rtr_arm_start();
	return ret;
}
module_init(msg_rtr_arm_init);

static void __exit msg_rtr_arm_exit(void)
{
	msg_rtr_arm_remove();
}
module_exit(msg_rtr_arm_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("NSN ARM-DSP DMA router driver");
